/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.messaging.exchange;

import java.net.URI;
import java.util.Map;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.RobustInOnly;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.jbi.messaging.exchange.impl.DefaultMessageExchangeFactory;
import org.ow2.petals.jbi.messaging.exchange.types.InOnlyImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOptionalOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.InOutImpl;
import org.ow2.petals.jbi.messaging.exchange.types.RobustInOnlyImpl;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.util.ServiceEndpointHelper;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This implementation creates {@link MessageExchangeWrapper} instances of {@link MessageExchange}.
 * @author Olivier Fabre - EBM WebSourcing
 */
public abstract class AbstractMessageExchangeFactory implements MessageExchangeFactory {

    protected final LoggingUtil log;

    private final ServiceEndpoint defaultEndpoint;

    private final QName defaultInterfaceName;

    private final QName defaultServiceName;

    private final ComponentContext context;

    private final Location consumerLocation;

    protected AbstractMessageExchangeFactory(final ServiceEndpoint targetEndpoint,
            final QName targetInterface, final QName targetService, final Location consumerLocation,
            final ComponentContext context, final Logger log) {
        assert context != null;
        assert log != null;
        assert consumerLocation != null;
        this.defaultEndpoint = targetEndpoint;
        this.defaultInterfaceName = targetInterface;
        this.defaultServiceName = targetService;
        this.context = context;
        this.log = new LoggingUtil(log);
        this.consumerLocation = consumerLocation;
    }

    @Override
    public InOnly createInOnlyExchange() throws MessagingException {
        this.log.call();

        return (InOnly) buildExchangeDecorator(MEPPatternConstants.IN_ONLY.value());
    }

    @Override
    public InOptionalOut createInOptionalOutExchange() throws MessagingException {
        this.log.call();

        return (InOptionalOut) buildExchangeDecorator(MEPPatternConstants.IN_OPTIONAL_OUT.value());
    }

    @Override
    public InOut createInOutExchange() throws MessagingException {
        this.log.call();

        return (InOut) buildExchangeDecorator(MEPPatternConstants.IN_OUT.value());
    }

    @Override
    public RobustInOnly createRobustInOnlyExchange() throws MessagingException {
        this.log.call();

        return (RobustInOnly) buildExchangeDecorator(MEPPatternConstants.ROBUST_IN_ONLY.value());
    }

    /**
     * TODO the exchange created has not MEP and the caller can't set it!!! we need to parse the WSDL then!
     */
    @Override
    public MessageExchange createExchange(final QName serviceName, QName operationName)
            throws MessagingException {
        this.log.call("Service Name : " + serviceName + " - operationName : " + operationName);
        assert serviceName != null;
        assert operationName != null;

        final ServiceEndpoint[] endpoints = this.context.getEndpointsForService(serviceName);
        if (endpoints.length == 0) {
            throw new MessagingException("No endpoint is registered for service k" + serviceName + "'");
        }

        if (this.defaultInterfaceName != null && !matches(endpoints, this.defaultInterfaceName)) { 
            throw new MessagingException("No endpoint for service '" + serviceName
                    + "' implements factory's default interface '" + this.defaultInterfaceName + "'");
        }

        // let's prepare an exception in case there is a problem
        final MessagingException e = new MessagingException(
                "No description for service '" + serviceName + "', can't find operation's MEP");
        Document doc = null;
        String endpointName = null;
        for (final ServiceEndpoint endpoint : endpoints) {
            try {
                doc = this.context.getEndpointDescriptor(endpoint);
                if (doc != null) {
                    endpointName = endpoint.getEndpointName();
                    break;
                } else {
                    this.log.warning("No description available for endpoint " + endpoint);
                }
            } catch (final JBIException e1) {
                e.addSuppressed(new MessagingException(
                        "Can't retrieve endpoint descriptor for endpoint '" + endpoint + "'", e1));
            }
        }

        if (doc == null) {
            throw e;
        }

        final Map<QName, MEPPatternConstants> meps = ServiceEndpointHelper
                .getOperationsMEPFromComponentDescription(serviceName, endpointName, doc, this.log.getLogger());

        final MEPPatternConstants mep = meps.get(operationName);
        if (mep == null) {
            throw new MessagingException("No operation defined for service '" + serviceName + "' named '"
                    + operationName + "', can't find operation's MEP");
        }

        final MessageExchangeWrapper messageExchangeDecorator = buildExchangeDecorator(mep.value());
        messageExchangeDecorator.setService(serviceName);
        messageExchangeDecorator.setOperation(operationName);

        return messageExchangeDecorator;
    }

    /**
     * @return <code>true</code> if one of the endpoint matches the interface
     */
    private static boolean matches(final ServiceEndpoint[] endpoints, final QName interfaceName) {
        for (final ServiceEndpoint endpoint : endpoints) {
            if (ServiceEndpointHelper.matches(endpoint, interfaceName)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public MessageExchange createExchange(final URI pattern) throws MessagingException {
        this.log.call(pattern);
        assert pattern != null;

        checkPatternIsJBISupportedPattern(pattern);

        return buildExchangeDecorator(pattern);
    }

    /**
     * Used to ensure the created decorator is coherent with the pattern set
     * 
     * Note: it important to test RobustInOnly before InOnly and InOptionalOut before InOut because of the inheritance
     * of the implementation classes!
     */
    private boolean mepIsValid(final MessageExchangeWrapper decorator) {
        if (decorator instanceof RobustInOnly) {
            return MEPPatternConstants.ROBUST_IN_ONLY.equals(decorator.getPattern());
        }
        if (decorator instanceof InOnly) {
            return MEPPatternConstants.IN_ONLY.equals(decorator.getPattern());
        }
        if (decorator instanceof InOptionalOut) {
            return MEPPatternConstants.IN_OPTIONAL_OUT.equals(decorator.getPattern());
        }
        if (decorator instanceof InOut) {
            return MEPPatternConstants.IN_OUT.equals(decorator.getPattern());
        }
        return true;
    }

    private void checkPatternIsJBISupportedPattern(final URI pattern) throws MessagingException {
        if (!MEPPatternConstants.IN_ONLY.equals(pattern) && !MEPPatternConstants.ROBUST_IN_ONLY.equals(pattern)
                && !MEPPatternConstants.IN_OUT.equals(pattern) && !MEPPatternConstants.IN_OPTIONAL_OUT.equals(pattern)) {
            throw new MessagingException("This Message Exchange Pattern is not recognized by JBI.");
        }
    }

    protected MessageExchangeWrapper buildExchangeDecorator(final URI pattern) throws MessagingException {
        this.log.call(pattern);
        assert pattern != null;

        final MessageExchangeWrapper exchange = createMessageExchangeDecoratorForPattern(buildExchange(pattern));

        assert mepIsValid(exchange);

        return exchange;
    }

    /**
     * This is responsible of building and configuring the exchange with the default values of the factory
     */
    protected PetalsMessageExchange buildExchange(final URI pattern) throws MessagingException {
        assert pattern != null;
        final PetalsMessageExchange msg = buildExchange(pattern, this.consumerLocation);

        msg.setEndpoint(defaultEndpoint);
        msg.setInterfaceName(defaultInterfaceName);
        msg.setService(defaultServiceName);

        return msg;
    }

    /**
     * Override to create a new {@link PetalsMessageExchange}.
     * 
     * @see DefaultMessageExchangeFactory for an example.
     */
    protected abstract PetalsMessageExchange buildExchange(URI pattern, Location consumerLocation)
            throws MessagingException;

    /**
     * This is responsible of creating a {@link MessageExchangeWrapper} from a {@link PetalsMessageExchange}.
     * 
     * Override, if needed, to customize
     */
    protected MessageExchangeWrapper createMessageExchangeDecoratorForPattern(final PetalsMessageExchange exchange) {
        final URI pattern = exchange.getPattern();
        if (isInOnlyPattern(pattern)) {
            return new InOnlyImpl(exchange);
        } else if (isRobustInOnlyPattern(pattern)) {
            return new RobustInOnlyImpl(exchange);
        } else if (isInOutPattern(pattern)) {
            return new InOutImpl(exchange);
        } else if (isInOptionalOutPattern(pattern)) {
            return new InOptionalOutImpl(exchange);
        } else {
            if (pattern != null) {
                this.log.warning("Creating an exchange with unknown pattern: " + pattern);
            }
            return new MessageExchangeWrapper(exchange);
        }
    }

    protected boolean isInOptionalOutPattern(final URI pattern) {
        return MEPPatternConstants.IN_OPTIONAL_OUT.equals(pattern);
    }

    protected boolean isInOutPattern(final URI pattern) {
        return MEPPatternConstants.IN_OUT.equals(pattern);
    }

    protected boolean isRobustInOnlyPattern(final URI pattern) {
        return MEPPatternConstants.ROBUST_IN_ONLY.equals(pattern);
    }

    protected boolean isInOnlyPattern(final URI pattern) {
        return MEPPatternConstants.IN_ONLY.equals(pattern);
    }
}
