/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.messaging.exchange.impl;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;

/**
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Gael BLONDELLE - EBM WebSourcing
 * @author Anass OUAZZANI - EBM WebSourcing
 * @author Victor Noël - Linagora
 */
public class MessageExchangeImpl implements PetalsMessageExchange, Cloneable {

    private static final long serialVersionUID = 429483387216034585L;

    private static final short SERIALIZE_ROLE_CONSUMER = 0;

    private static final short SERIALIZE_ROLE_PROVIDER = 1;

    private static final short SERIALIZE_STATUS_ACTIVE = 0;

    private static final short SERIALIZE_STATUS_DONE = 1;

    private static final short SERIALIZE_STATUS_ERROR = 2;

    private final Location consumerLocation;

    private PetalsServiceEndpoint providerEndpoint;

    private ServiceEndpoint endpoint;

    private Exception error;

    private final String exchangeId;

    private QName interfaceName;

    private Map<String, NormalizedMessage> messages = new HashMap<String, NormalizedMessage>();

    private QName operation;

    private final URI pattern;

    private Map<String, Object> properties = new HashMap<String, Object>();

    private QName service;

    /**
     * Contains the termination reason (to ease understanding errors)
     */
    private String termination = null;

    private transient Role role = Role.CONSUMER;

    private transient ExchangeStatus status = ExchangeStatus.ACTIVE;

    private boolean doChecks = true;

    public MessageExchangeImpl(final String exchangeId, final URI pattern, final Location consumerLocation) {
        this.exchangeId = exchangeId;
        this.pattern = pattern;
        this.consumerLocation = consumerLocation;
    }

    @Override
    public Fault createFault() throws MessagingException {
        this.checkIsActive();
        // note: we don't check if it is in a valid state, just that it can create a fault
        // setFault will check if we are in the valid state!
        this.checkCallerIsAllowedToCreateFault();
        return new FaultImpl();
    }

    private void checkCallerIsAllowedToCreateFault() throws MessagingException {
        if (!doChecks()) {
            return;
        }

        if ((isRobustInOnly() && isProvider()) || (isInOut() && isProvider())
                || isInOptionalOut()) {
            return;
        }

        throw new MessagingException(
                "The MessageExchange state does not allow this operation: create FAULT on: " + getRoleStatusAndState());
    }

    @Override
    public NormalizedMessage createMessage() throws MessagingException {
        this.checkIsActive();
        return new NormalizedMessageImpl();
    }

    @Override
    public ServiceEndpoint getEndpoint() {
        // according to JBI, if an endpoint was specified, then it must be the one returned
        return this.endpoint != null ? this.endpoint : this.providerEndpoint;
    }

    @Override
    public Exception getError() {
        return this.error;
    }

    @Override
    public String getExchangeId() {
        return this.exchangeId;
    }

    @Override
    public Fault getFault() {
        return (Fault) this.getMessage(FAULT_MSG);
    }

    @Override
    public QName getInterfaceName() {
        // the currently used provider endpoint always takes precedence over what was specified
        // in case there is multiple interfaces in it, if the one specified is present, it is used, if not, the first
        // one is taken
        if (this.providerEndpoint != null) {
            final QName[] interfaces = this.providerEndpoint.getInterfaces();
            if (this.interfaceName != null) {
                for (final QName name : interfaces) {
                    if (this.interfaceName.equals(name)) {
                        return this.interfaceName;
                    }
                }
            }
            // by default let's use the first one
            if (interfaces.length > 0) {
                return interfaces[0];
            } else {
                return this.interfaceName;
            }
        } else {
            return this.interfaceName;
        }
    }

    /**
     * get the message associated to the reference.
     * 
     * It is <b>Not</b> Case Sensitive.
     */
    @Override
    public NormalizedMessage getMessage(final String name) {
        if (name != null) {
            return this.messages.get(name.toLowerCase());
        } else {
            return null;
        }
    }

    @Override
    public Map<String, NormalizedMessage> getMessages() {
        return this.messages;
    }

    @Override
    public QName getOperation() {
        return this.operation;
    }

    @Override
    public URI getPattern() {
        return this.pattern;
    }

    @Override
    public Object getProperty(final String name) {
        return this.properties.get(name);
    }

    @Override
    public Set<?> getPropertyNames() {
        return this.properties.keySet();
    }

    @Override
    public Role getRole() {
        return this.role;
    }

    @Override
    public QName getService() {
        // the currently used provider endpoint always takes precedence over what was specified
        return this.providerEndpoint != null ? this.providerEndpoint.getServiceName() : this.service;
    }

    @Override
    public ExchangeStatus getStatus() {
        return this.status;
    }

    @Override
    public String getTermination() {
        return this.termination;
    }

    @Override
    public boolean isTransacted() {
        return false;
    }

    @Override
    public Location getConsumerLocation() {
        return this.consumerLocation;
    }

    @Override
    public void setElectedProviderEndpoint(final PetalsServiceEndpoint endpoint) {
        this.providerEndpoint = endpoint;
    }

    @Override
    public PetalsServiceEndpoint getElectedProviderEndpoint() {
        return this.providerEndpoint;
    }

    @Override
    public void setEndpoint(final ServiceEndpoint endpoint) {
        this.endpoint = endpoint;
    }

    // //////////////////////////////////////////
    // specific methods
    // //////////////////////////////////////////

    @Override
    public void setError(final Exception error) {
        if (error != null) {
            this.error = error;
            this.status = ExchangeStatus.ERROR;
        }
    }

    @Override
    public void setFault(final Fault fault) throws MessagingException {
        this.setMessage(fault, FAULT_MSG);
    }

    @Override
    public void setDoChecks(final boolean doChecks) {
        this.doChecks = doChecks;
    }

    private boolean doChecks() {
        // we need the property hack because wrappers of MessageExchangeImpl potentially can't see the class or
        // are in different classloaders.
        final Object ecDontDoChecks = getProperty(PROPERTY_DONT_DO_CHECKS);
        return doChecks && ecDontDoChecks == null || (ecDontDoChecks instanceof Boolean && !((boolean) ecDontDoChecks));
    }

    private void checkCallerIsAllowedToSetFault() throws MessagingException {

        if (isRobustInOnly() && isProvider()) {
            return;
        }

        // we can set a Fault if there is an OUT already (but it will be removed!)
        if ((isInOut() || isInOptionalOut()) && isProvider()) {
            return;
        }

        // only if there is already an OUT set it is valid as a consumer
        if (isInOptionalOut() && isConsumer() && hasOut()) {
            return;
        }

        // all other cases are incorrect
        throw new MessagingException("The MessageExchange state does not allow this operation: setting FAULT on: "
                + getRoleStatusAndState());
    }

    private void checkCallerIsAllowedToSetIn() throws MessagingException {

        if (isConsumer()) {
            if (isInOnly()) {
                return;
            }

            if ((isInOut() || isInOptionalOut()) && !hasOut() && !hasFault()) {
                return;
            }

            if (isRobustInOnly() && !hasFault()) {
                return;
            }
        }

        // all other cases are incorrect
        throw new MessagingException(
                "The MessageExchange state does not allow this operation: setting IN on: " + getRoleStatusAndState());
    }

    private void checkCallerIsAllowedToSetOut() throws MessagingException {

        // we can set an out if there is a fault (but it will be removed!)
        if (isProvider() && (isInOut() || isInOptionalOut())) {
            return;
        }

        // all other cases are incorrect
        throw new MessagingException(
                "The MessageExchange state does not allow this operation: setting OUT on: "
                        + getRoleStatusAndState());
    }

    private boolean isInOptionalOut() {
        return MEPPatternConstants.IN_OPTIONAL_OUT.equals(this.pattern);
    }

    private boolean isInOnly() {
        return MEPPatternConstants.IN_ONLY.equals(this.pattern);
    }

    private boolean isInOut() {
        return MEPPatternConstants.IN_OUT.equals(this.pattern);
    }

    private boolean isRobustInOnly() {
        return MEPPatternConstants.ROBUST_IN_ONLY.equals(this.pattern);
    }
    
    @Override
    public boolean hasIn() {
        return this.messages.containsKey(IN_MSG);
    }

    @Override
    public boolean hasOut() {
        return this.messages.containsKey(OUT_MSG);
    }

    @Override
    public boolean hasFault() {
        return this.messages.containsKey(FAULT_MSG);
    }

    @Override
    public boolean isConsumer() {
        return Role.CONSUMER.equals(this.role);
    }

    @Override
    public boolean isProvider() {
        return Role.PROVIDER.equals(this.role);
    }

    @Override
    public boolean isActive() {
        return ExchangeStatus.ACTIVE.equals(this.status);
    }

    @Override
    public boolean isDone() {
        return ExchangeStatus.DONE.equals(this.status);
    }

    @Override
    public boolean isError() {
        return ExchangeStatus.ERROR.equals(this.status);
    }

    private void checkIsActive() throws MessagingException {
        if (!doChecks()) {
            return;
        }

        this.checkIsNotTerminated();

        if (!isActive()) {
            throw new MessagingException(
                    "The MessageExchange state does not allow this operation: status is not Active");
        }
    }

    /**
     * See {@link PetalsMessageExchange#getTermination()} documentation to understand why this check is needed.
     */
    private void checkIsNotTerminated() throws MessagingException {
        if (this.termination != null) {
            throw new MessagingException(
                    "The MessageExchange state does not allow this operation: exchange is terminated (reason: "
                            + this.termination + ")");
        }
    }

    private void checkCallerIsAllowedToSetMessage(final String name) throws MessagingException {
        if (!doChecks()) {
            return;
        }

        if (this.pattern == null) {
            throw new MessagingException("The MEP is not defined.");
        }

        if (this.role == null) {
            throw new MessagingException("The Role is not defined.");
        }

        if (IN_MSG.equalsIgnoreCase(name)) {
            checkCallerIsAllowedToSetIn();
            return;
        }

        if (OUT_MSG.equalsIgnoreCase(name)) {
            checkCallerIsAllowedToSetOut();
            return;
        }

        if (FAULT_MSG.equalsIgnoreCase(name)) {
            checkCallerIsAllowedToSetFault();
            return;
        }

        // TODO do we really want to fail if another name is used?
        throw new MessagingException("The MessageExchange state does not allow this operation: setting " + name
                + " on: " + getRoleStatusAndState());
    }

    private void checkCallerIsAllowedToSetStatus(final ExchangeStatus status) throws MessagingException {
        if (!doChecks()) {
            return;
        }

        this.checkIsNotTerminated();

        // Note that it should be ok for someone to set a message to done, then to active, then to add a fault, then to
        // remove it, then to set it to done again or things like that!
        if (ExchangeStatus.DONE.equals(status)) {
            if (isConsumer()) {
                // valid cases
                if ((isRobustInOnly() && hasFault())
                        || (isInOut() && (hasOut() || hasFault()))
                        || (isInOptionalOut() && (hasOut() ^ hasFault()))) {
                    return;
                }
            } else {
                // valid cases
                if (isInOnly()
                        || (isRobustInOnly() && !hasFault())
                        || (isInOptionalOut() && !hasFault() && !hasOut())
                        || (isInOptionalOut() && hasFault() && hasOut())) {
                    return;
                }
            }
        } else {
            // for ERROR status, there is no reason to fail if error is set (and the setError method does not throw the
            // usual MessagingException to indicate that such problem could arise)
            // for ACTIVE status, anyway, if the exchange is finished then it would have been terminated
            // other cases are ok!
            return;
        }
        // all other cases are wrong
        throw new MessagingException("The MessageExchange state does not allow this operation: setting status "
                + status + " on: " + getRoleStatusAndState());
    }

    /**
     * Used to populate errors in checks.
     */
    private String getRoleStatusAndState() {
        return "Status: " + this.status + ", Role: " + this.role + ", Pattern: "
                + MEPPatternConstants.fromURI(this.pattern).name() + (hasIn() ? ", hasInMsg" : "")
                + (hasOut() ? ", hasOutMsg" : "") + (hasFault() ? ", hasFaultMsg" : "");
    }

    @Override
    public void setInterfaceName(final QName interfaceName) {
        // TODO should we forbid setting interface if an endpoint has already been selected?
        this.interfaceName = interfaceName;
    }

    /**
     * Set the specified message. Check the availability to set the message with the following rules:
     * <ul>
     * <li>CONSUMER Role <-> in reference or PROVIDER Role <-> out reference</li>
     * <li>JBI predefined MEP <-> in or out reference</li>
     * </ul>
     * This is <b>not</b> case sensitive
     */
    @Override
    public void setMessage(final NormalizedMessage msg, final String name) throws MessagingException {
        if (msg == null) {
            throw new MessagingException("NormalizedMessage must be non null.");
        }
        if ((name == null) || (name.trim().length() == 0)) {
            throw new MessagingException("The message reference must be non null and non empty.");
        }

        this.checkIsActive();
        this.checkCallerIsAllowedToSetMessage(name);

        if (OUT_MSG.equalsIgnoreCase(name)) {
            // when setting an OUT, we are a provider and the message is InOut or InOptionalOut
            // in both case there can't be both an OUT and a Fault at the same time,
            // so the OUT will replace the Fault if there is one
            this.messages.remove(FAULT_MSG);
        } else if (FAULT_MSG.equalsIgnoreCase(name) && !(isInOptionalOut() && hasOut() && isConsumer())) {
            // When setting a Fault, except if we are setting it as a response to an Out in a InOptionalOut pattern,
            // then we can't have at the same time a Fault and an Out, so the Fault will replace the OUT if there is one
            this.messages.remove(OUT_MSG);
        }

        // no exception was thrown, set the message
        this.messages.put(name.toLowerCase(), msg);
    }

    @Override
    public void setOperation(final QName name) {
        this.operation = name;
    }

    @Override
    public void setProperty(final String name, final Object obj) {
        if (obj != null) {
            this.properties.put(name, obj);
        } else {
            this.properties.remove(name);
        }
    }

    @Override
    public void setRole(final Role role) {
        this.role = role;
    }

    @Override
    public void setService(final QName service) {
        // TODO should we forbid setting service if an endpoint has already been selected?
        this.service = service;
    }

    @Override
    public void setStatus(final ExchangeStatus status) throws MessagingException {
        checkCallerIsAllowedToSetStatus(status);
        this.status = status;
    }

    @Override
    public void setTermination(final String termination) {
        this.termination = termination;
    }

    @Override
    public String toString() {
        return "MessageExchange@" + this.getExchangeId();
    }

    @Override
    public boolean isNewExchange() {
        return isActive() && !hasFault() && !hasOut();
    }

    /**
     * Used to deserialize Status and Role objects
     * 
     * @param s
     * @throws IOException
     */
    private void readObject(final ObjectInputStream s) throws IOException {
        // Role deserialization
        switch (s.readShort()) {
            case 0:
                this.role = MessageExchange.Role.CONSUMER;
                break;
            case 1:
                this.role = MessageExchange.Role.PROVIDER;
                break;
            default:
                break;
        }

        // Status deserialization
        switch (s.readShort()) {
            case 0:
                this.status = ExchangeStatus.ACTIVE;
                break;
            case 1:
                this.status = ExchangeStatus.DONE;
                break;
            case 2:
                this.status = ExchangeStatus.ERROR;
                break;
            default:
                break;
        }

        try {
            s.defaultReadObject();
        } catch (final ClassNotFoundException e) {
            throw new IOException(e);
        }
    }

    /**
     * Used to serialize Status and Role objects
     * 
     * @param s
     * @throws IOException
     */
    private void writeObject(final ObjectOutputStream s) throws IOException {
        // Role serialization
        if (isConsumer()) {
            s.writeShort(SERIALIZE_ROLE_CONSUMER);
        } else {
            s.writeShort(SERIALIZE_ROLE_PROVIDER);
        }

        // Status serialization
        if (ExchangeStatus.ACTIVE.equals(this.status)) {
            s.writeShort(SERIALIZE_STATUS_ACTIVE);
        } else if (ExchangeStatus.DONE.equals(this.status)) {
            s.writeShort(SERIALIZE_STATUS_DONE);
        } else {
            s.writeShort(SERIALIZE_STATUS_ERROR);
        }
        s.defaultWriteObject();
    }
}
