/**
 * Copyright (c) 2016-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.servicedesc.endpoint.util;

import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.easywsdl.extensions.wsdl4complexwsdl.WSDL4ComplexWsdlReaders;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.WSDL4ComplexWsdlReader;
import org.ow2.easywsdl.wsdl.api.Binding;
import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.Operation;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfOperation.MEPPatternConstants;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.w3c.dom.Document;

/**
 * Some helpers pertaining to manipulating {@link ServiceEndpoint} and their descriptions.
 * 
 * @author vnoel
 *
 */
public class ServiceEndpointHelper {

    public static QName[] getInterfacesForEndpoint(final QName serviceName, final String endpointName,
            final Document doc, final Logger logger) {
        final QName[] interfaces;
        if (doc != null) {
            final List<QName> extractedInterfaces = getInterfacesFromComponentDescription(serviceName, endpointName,
                    doc, logger);
            if (!extractedInterfaces.isEmpty()) {
                interfaces = extractedInterfaces.toArray(new QName[extractedInterfaces.size()]);
            } else {
                final QName defaultInterface = getDefaultInterface(serviceName);
                logger.warning("No interface has been found in the component service description, set default to "
                        + defaultInterface);
                interfaces = new QName[] { defaultInterface };
            }
        } else {
            final QName defaultInterface = getDefaultInterface(serviceName);
            logger.warning("No description has been found for service '" + serviceName + "', endpoint name '"
                    + endpointName + "'. Interface set default to " + defaultInterface);
            interfaces = new QName[] { defaultInterface };
        }
        return interfaces;
    }

    /**
     * Note: return <code>true</code> if interfaces is empty (see PETALSESBCONT-398)
     */
    public static boolean matches(final ServiceEndpoint endpoint, final QName name) {
        if (endpoint.getInterfaces().length == 0) {
            return true;
        }

        for (final QName itf : endpoint.getInterfaces()) {
            if (name.equals(itf)) {
                return true;
            }
        }

        return false;
    }

    public static Map<QName, MEPPatternConstants> getOperationsMEPFromComponentDescription(final QName serviceName,
            final String endpointName, final Document doc, final Logger logger) {
        final Map<QName, MEPPatternConstants> meps = new HashMap<>();

        try {
            WSDL4ComplexWsdlReader wsdl4ComplexWsdlReader = null;
            try {
                wsdl4ComplexWsdlReader = WSDL4ComplexWsdlReaders.takeWSDL4ComplexWsdlReader();
                Description wsdlDescription = wsdl4ComplexWsdlReader.read(doc);

                if (wsdlDescription != null) {
                    final InterfaceType interfaceType = getInterfaceFromService(serviceName, endpointName,
                            wsdlDescription, logger);

                    if (interfaceType != null && interfaceType.getQName() != null) {
                        for (final Operation operation : interfaceType.getOperations()) {
                            if (operation.getQName() != null && operation.getPattern() != null) {
                                meps.put(operation.getQName(), operation.getPattern());
                            } else {
                                logger.warning("Invalid data in component description for operation '"
                                        + operation.getQName() + "' of service '" + serviceName + "', MEP is '"
                                        + operation.getPattern() + "'");
                            }
                        }
                    } else {
                        logger.warning("No interface in component description for service '" + serviceName
                                + "' and endpoint '" + endpointName + "', MEPs not gathered");
                    }
                } else {
                    logger.warning("No component description for service '" + serviceName + "', MEPs not gathered");
                }
            } catch (final WSDLException e) {
                // it's thrown by ss.getInterface(), there is no alternative
                logger.log(Level.WARNING,
                        "Error while getting interface for service '" + serviceName + "', MEPs not gathered", e);
            } finally {
                if (wsdl4ComplexWsdlReader != null) {
                    WSDL4ComplexWsdlReaders.releaseWSDL4ComplexWsdlReader(wsdl4ComplexWsdlReader);
                }
            }
        } catch (final URISyntaxException e) {
            logger.log(Level.WARNING, "Can't read component description for service '" + serviceName + "'", e);
        }

        return meps;
    }

    /**
     * in WSDL2, the service has an interface defined, but in WSDL1, we are not totally assured that getting interface
     * on the service will succeed, so it is safer to check directly the binding. We fallback on the service's interface
     * else.
     */
    private static InterfaceType getInterfaceFromService(final QName serviceName, final String endpointName,
            final Description description, final Logger logger) throws WSDLException {
        final Service ss = description.getService(serviceName);
        if (ss != null) {
            final Endpoint endpoint = ss.getEndpoint(endpointName);
            if (endpoint != null) {
                final Binding binding = endpoint.getBinding();
                if (binding != null && binding.getInterface() != null) {
                    return binding.getInterface();
                } else {
                    logger.config("No valid binding for endpoint '" + endpointName
                            + "' in component description for service '" + serviceName
                            + "', using service's interface");
                    return ss.getInterface();
                }
            } else {
                logger.config("No endpoint '" + endpointName + "' in component description for service '" + serviceName
                        + "', using service's interface");
                return ss.getInterface();
            }
        } else {
            logger.config("No service '" + serviceName + "' in component description");
            return null;
        }
    }

    private static List<QName> getInterfacesFromComponentDescription(final QName serviceName, final String endpointName,
            final Document doc, final Logger logger) {
        final List<QName> result = new ArrayList<>();
        try {
            WSDL4ComplexWsdlReader wsdl4ComplexWsdlReader = null;
            try {
                wsdl4ComplexWsdlReader = WSDL4ComplexWsdlReaders.takeWSDL4ComplexWsdlReader();
                Description wsdlDescription = wsdl4ComplexWsdlReader.read(doc);

                if (wsdlDescription != null) {
                    final InterfaceType interfaceType = getInterfaceFromService(serviceName, endpointName,
                            wsdlDescription, logger);
                    if (interfaceType != null && interfaceType.getQName() != null) {
                        // TODO also add super interfaces!
                        result.add(interfaceType.getQName());
                    } else {
                        logger.warning(
                                "No binding for endpoint '" + endpointName + "' in component description for service '"
                                        + serviceName + "' and service has no interface defined");
                    }
                } else {
                    logger.warning("No component description for service '" + serviceName + "', interface not set");
                }
            } catch (final WSDLException e) {
                // it's thrown by ss.getInterface(), there is no alternative
                logger.log(Level.WARNING,
                        "Error while getting interface for service '" + serviceName + "', interface not set", e);
            } finally {
                if (wsdl4ComplexWsdlReader != null) {
                    WSDL4ComplexWsdlReaders.releaseWSDL4ComplexWsdlReader(wsdl4ComplexWsdlReader);
                }
            }
        } catch (final URISyntaxException e) {
            logger.log(Level.WARNING, "Can't read component description for service '" + serviceName + "'", e);
        }
        return result;
    }

    private static QName getDefaultInterface(final QName serviceName) {
        return new QName(serviceName.getNamespaceURI(),
                PetalsServiceEndpoint.UNRESOLVED_INTERFACE + serviceName.getLocalPart());
    }
}
