/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2017 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.messaging.exchange;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.URISyntaxException;
import java.net.URL;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.security.auth.Subject;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamSource;

import org.custommonkey.xmlunit.Diff;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.jbi.messaging.exchange.impl.NormalizedMessageImpl;
import org.ow2.petals.jbi.xml.BytesSource;
import org.xml.sax.InputSource;

import com.ebmwebsourcing.easycommons.xml.SourceHelper;

/**
 * Test of the NormalizedMessageImpl.
 * @author ddesjardins - EBM WebSourcing
 */
public class NormalizedMessageImplTest {

    @Rule
    public TemporaryFolder tempFolder = new TemporaryFolder();

    private NormalizedMessageImpl normalizedMessageImpl;

    private FileDataSource source;

    @Before
    public void before() {

        final URL sourceUrl = Thread.currentThread().getContextClassLoader()
                .getResource("filesource.txt");
        try {
            source = new FileDataSource(new File(sourceUrl.toURI()));
        } catch (URISyntaxException e) {
            e.printStackTrace();
        }
        normalizedMessageImpl = new NormalizedMessageImpl();
    }

    /**
     * Check that no error occurs adding an attachment
     */
    @Test
    public void testAddAttachment() throws IOException, MessagingException {
        normalizedMessageImpl.addAttachment("test", new DataHandler(source));
    }

    /**
     * Check that no error occurs getting an attachment previously added
     */
    @Test
    public void testGetAttachment() throws IOException, MessagingException {
        testAddAttachment();
        DataHandler dataHandler = normalizedMessageImpl.getAttachment("test");
        String content = (String) dataHandler.getContent();
        assertTrue(content.startsWith("# Copyright (c) 2011-2012 EBM WebSourcing, 2012-20", 2));
        assertTrue(content.endsWith("File used to test the normalized message"));
    }

    /**
     * Check that no error occurs getting an attachment not previously added,
     * returned value is <code>null</code>
     */
    @Test
    public void testGetNotExistingAttachment() throws IOException, MessagingException {
        testAddAttachment();
        assertNull(normalizedMessageImpl.getAttachment("foo"));
    }

    /**
     * Check that the content can be retrieved
     * 
     * @throws TransformerException
     */
    @Test
    public void testGetContent() throws Exception {
        normalizedMessageImpl = new NormalizedMessageImpl();
        final URL descriptorUrl = Thread.currentThread().getContextClassLoader()
                .getResource("test-content.xml");
        Source xml = new StreamSource(new File(descriptorUrl.toURI()));
        normalizedMessageImpl.setContent(xml);

        final File normalizedMsgFile = this.tempFolder.newFile("normalized-message");
        try (final OutputStream normalizedMsgFos = new FileOutputStream(normalizedMsgFile)) {
            try (final ObjectOutputStream normalizedMsgOos = new ObjectOutputStream(normalizedMsgFos)) {
                normalizedMsgOos.writeObject(this.normalizedMessageImpl);
            }
        }

        try (final InputStream normalizedMsgFis = new FileInputStream(normalizedMsgFile)) {
            try (final ObjectInputStream normalizedMsgOis = new ObjectInputStream(normalizedMsgFis)) {
                final Object obj = normalizedMsgOis.readObject();
                assertNotNull(obj);
                assertTrue(obj instanceof NormalizedMessageImpl);
                this.normalizedMessageImpl = (NormalizedMessageImpl) obj;
                final Source source = this.normalizedMessageImpl.getContent();
                assertNotNull(source);
                // default implementation used in the serializer
                assertTrue(source instanceof BytesSource);
                final InputSource inputSource2 = SourceHelper.toInputSource(source);
                final Diff xmlDiff = new Diff(new InputSource(new FileInputStream(new File(
                        descriptorUrl.toURI()))), inputSource2);
                assertTrue(xmlDiff.similar());
            }
        }
    }

    /**
     * Check that no error occurs getting name of attachments
     */
    @Test
    public void testGetAttachmentNames() throws IOException, MessagingException {
        testAddAttachment();
        assertEquals(normalizedMessageImpl.getAttachmentNames().iterator().next(), "test");
    }

    /**
     * Check that no error occurs removing an attachment previously added.
     */
    @Test
    public void testRemoveAttachment() throws IOException, MessagingException {
        testAddAttachment();
        normalizedMessageImpl.removeAttachment("test");
        assertNull(normalizedMessageImpl.getAttachment("test"));
    }

    /**
     * Check that an error occurs removing an attachment that was not previously
     * added.
     */
    @Test(expected = MessagingException.class)
    public void testRemoveNotExistingAttachment() throws IOException, MessagingException {
        testAddAttachment();
        normalizedMessageImpl.removeAttachment("foo");
    }

    /**
     * Check that a property is correctly set into the {@link NormalizedMessage}
     * , and its value is unmodified
     */
    @Test
    public void testGetPropertyNames() throws IOException {
        normalizedMessageImpl = new NormalizedMessageImpl();
        normalizedMessageImpl.setProperty("test", "value");
        assertEquals(normalizedMessageImpl.getPropertyNames().iterator().next(), "test");
    }

    /**
     * Check that a security subject is correctly set into the
     * {@link NormalizedMessage}, and its value is unmodified
     */
    @Test
    public void testGetSecuritySubject() {
        Subject subject = new Subject();
        normalizedMessageImpl.setSecuritySubject(subject);
        assertEquals(normalizedMessageImpl.getSecuritySubject(), subject);
    }

    /**
     * Check that a property value can be overriden
     */
    @Test
    public void testSetProperty() {
        normalizedMessageImpl.setProperty("test", mock(DataHandler.class));
        normalizedMessageImpl.setProperty("test", null);
        assertNull(normalizedMessageImpl.getProperty("test"));
    }

}
